#!/usr/bin/env python3
# --------------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
# --------------------------------------------------------------------------------------
from abc import ABC
from abc import abstractmethod
from asyncio import as_completed
from asyncio import gather
from asyncio import sleep
from contextlib import asynccontextmanager
from functools import partial
from itertools import groupby
from itertools import starmap
from typing import Callable
from typing import Coroutine
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from uuid import UUID

from aiohttp import ClientSession
from aiohttp import TCPConnector
from more_itertools import chunked
from pydantic import AnyHttpUrl
from ramodels.base import RABase
from tqdm import tqdm


class ModelClientBase(ABC):
    def __init__(
        self,
        base_url: AnyHttpUrl,
        chunk_size: int = 100,
        saml_token: Optional[UUID] = None,
        session_factory: Callable[[], ClientSession] = partial(
            ClientSession, connector=TCPConnector(limit=20)
        ),
    ):
        # connection logic
        self._base_url = base_url
        self._chunk_size = chunk_size
        if saml_token:
            session_factory = partial(
                session_factory, headers={"SESSION": str(saml_token)}
            )
        self._session_factory = session_factory

        self._session: Optional[ClientSession] = None

    @asynccontextmanager
    async def context(self):
        try:
            async with self._session_factory() as session:
                self._session = session
                await self.__check_if_server_online()

                yield
        finally:
            self._session = None

    async def _verify_session(self) -> ClientSession:
        if self._session is None:
            raise Exception("Need to initialize client session!")
        return self._session

    async def __check_if_server_online(self, attempts=100, delay=1) -> None:
        """Check if backend is online.

        :param attempts: Number of repeats
        :param delay: Number of sleeps in-between
        :return:
        """
        session = await self._verify_session()

        async def check_endpoint(url, response):
            for _ in range(attempts):
                try:
                    resp = await session.get(url)
                    resp.raise_for_status()
                    if response in await resp.json():
                        return
                    raise Exception("Invalid response")
                except Exception as exp:
                    print(exp)
                    await sleep(delay)
            raise Exception("Unable to connect")

        healthcheck_tuples = self._get_healthcheck_tuples()
        healthcheck_tuples = [
            (self._base_url + subpath, response)
            for subpath, response in healthcheck_tuples
        ]
        tasks = starmap(check_endpoint, healthcheck_tuples)
        await gather(*tasks)

    async def _post_to_backend(
        self, current_type: Type[RABase], data: Iterable[RABase]
    ):
        """
        wrapper allows passing list of mox objs, for individual posting
        :param current_type:
        :param data:
        :return:
        """
        await self._verify_session()
        await gather(
            *map(
                lambda obj: self._post_single_to_backend(
                    current_type=current_type, obj=obj
                ),
                data,
            )
        )

    async def __submit(self, data: Iterable[RABase]):
        """
        maps the object appropriately to either MO or LoRa

        :param data: An iterable of objects of the *same* type!
        :return:
        """

        data = list(data)
        current_type = type(data[0])

        assert all([isinstance(obj, current_type) for obj in data])

        if current_type in self._get_path_map():
            await self._post_to_backend(current_type, data)
            return

        raise TypeError(f"unknown type: {current_type}")

    async def _submit_payloads(self, objs: Iterable[RABase], disable_progressbar=False):
        objs = list(objs)
        groups = groupby(objs, lambda x: type(x).__name__)
        chunked_groups: List[Tuple[str, Iterable[List[RABase]]]] = [
            (type_name, list(chunked(objs, self._chunk_size)))
            for type_name, objs in groups
        ]
        chunked_tasks: List[Tuple[str, List[Coroutine]]] = [
            (type_name, list(map(self.__submit, chunks)))
            for type_name, chunks in chunked_groups
        ]
        if not chunked_tasks or all([not tasks for _, tasks in chunked_tasks]):
            return
        for key, tasks in chunked_tasks:
            for f in tqdm(
                as_completed(tasks),
                total=len(tasks),
                unit="chunk",
                desc=key,
                disable=disable_progressbar,
            ):
                await f

    @abstractmethod
    def _get_healthcheck_tuples(self) -> List[Tuple[str, str]]:
        pass

    @abstractmethod
    def _get_path_map(self) -> Dict[RABase, str]:
        pass

    @abstractmethod
    async def _post_single_to_backend(self, current_type: Type[RABase], obj: RABase):
        pass



#!/usr/bin/env python3
# --------------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
# --------------------------------------------------------------------------------------
from asyncio import run
from typing import Dict
from typing import Iterable
from typing import List
from typing import Tuple
from typing import Type
from uuid import UUID

from pydantic import AnyHttpUrl
from ramodels.base import RABase
from ramodels.mo import Address
from ramodels.mo import Employee
from ramodels.mo import Engagement
from ramodels.mo import EngagementAssociation
from ramodels.mo import Manager
from ramodels.mo import OrganisationUnit

from raclients.modelclientbase import ModelClientBase
from raclients.util import uuid_to_str

MoObj = Type[RABase]


class ModelClient(ModelClientBase):
    __mo_path_map = {
        OrganisationUnit: "/service/ou/create",
        Employee: "/service/e/create",
        Engagement: "/service/details/create",
        EngagementAssociation: "/service/details/create",
        Manager: "/service/details/create",
        Address: "/service/details/create",
    }

    def __init__(self, base_url: AnyHttpUrl = "http://localhost:5000", *args, **kwargs):
        super().__init__(base_url, *args, **kwargs)

    def _get_healthcheck_tuples(self) -> List[Tuple[str, str]]:
        return [("/version", "mo_version")]

    def _get_path_map(self) -> Dict[RABase, str]:
        return self.__mo_path_map

    async def _post_single_to_backend(self, current_type: Type[MoObj], obj: MoObj):
        session = await self._verify_session()

        async with session.post(
            self._base_url + self.__mo_path_map[current_type],
            json=uuid_to_str(obj.dict(by_alias=True)),
        ) as response:
            response.raise_for_status()

    async def load_mo_objs(
        self, objs: Iterable[MoObj], disable_progressbar: bool = False
    ):
        """
        lazy init client session to ensure created within async context
        :param objs:
        :param disable_progressbar:
        :return:
        """
        await self._submit_payloads(objs, disable_progressbar=disable_progressbar)


if __name__ == "__main__":

    async def main():
        client = ModelClient()
        async with client.context():
            await client.load_mo_objs(
                [
                    Employee.from_simplified_fields(
                        uuid=UUID("456362c4-0ee4-4e5e-a72c-751239745e64"),
                        name="test_org_name",
                        user_key="test_org_user_key",
                    )
                ]
            )

    run(main())



#!/usr/bin/env python3
# --------------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
# --------------------------------------------------------------------------------------
from asyncio import run
from typing import Dict
from typing import Iterable
from typing import List
from typing import Tuple
from typing import Type

from pydantic import AnyHttpUrl
from ramodels.base import RABase
from ramodels.lora import Facet
from ramodels.lora import Klasse
from ramodels.lora import Organisation

from raclients.modelclientbase import ModelClientBase
from raclients.util import uuid_to_str

LoraObj = Type[RABase]


class ModelClient(ModelClientBase):
    __mox_path_map = {
        Organisation: "/organisation/organisation",
        Facet: "/klassifikation/facet",
        Klasse: "/klassifikation/klasse",
    }

    def __init__(self, base_url: AnyHttpUrl = "http://localhost:8080", *args, **kwargs):
        super().__init__(base_url, *args, **kwargs)

    def _get_healthcheck_tuples(self) -> List[Tuple[str, str]]:
        return [("/version", "lora_version")]

    def _get_path_map(self) -> Dict[RABase, str]:
        return self.__mox_path_map

    async def _post_single_to_backend(self, current_type: Type[LoraObj], obj: LoraObj):
        """

        :param current_type: Redundant, only pass it because we already have it
        :param obj:
        :return:
        """
        session = await self._verify_session()

        uuid = obj.uuid
        # TODO, PENDING: https://github.com/samuelcolvin/pydantic/pull/2231
        # for now, uuid is included, and has to be excluded when converted to json
        jsonified = uuid_to_str(
            obj.dict(by_alias=True, exclude={"uuid"}, exclude_none=True)
        )
        generic_url = self._base_url + self.__mox_path_map[current_type]
        if uuid is None:  # post
            async with session.post(
                generic_url,
                json=jsonified,
            ) as response:
                response.raise_for_status()
        else:  # put
            async with session.put(
                generic_url + f"/{uuid}",
                json=jsonified,
            ) as response:
                response.raise_for_status()

    async def load_lora_objs(
        self, objs: Iterable[LoraObj], disable_progressbar: bool = False
    ):
        """
        lazy init client session to ensure created within async context
        :param objs:
        :param disable_progressbar:
        :return:
        """
        await self._submit_payloads(objs, disable_progressbar=disable_progressbar)


if __name__ == "__main__":

    async def main():
        client = ModelClient(base_url="https://morademo.magenta.dk")
        async with client.context():
            pass
            # await client.load_lora_objs(
            #    [
            #        Organisation.from_simplified_fields(
            #            uuid=UUID("456362c4-0ee4-4e5e-a72c-751239745e64"),
            #            name="test_org_name",
            #            user_key="test_org_user_key",
            #        )
            #    ]
            # )

    run(main())
